#!/usr/bin/env node
const cwd = process.cwd()
const git = require('simple-git')(cwd)
const args = require('minimist')(process.argv.slice(2))
const { prompt } = require('inquirer')
const { green, redBright } = require('chalk')
const figlet = require('figlet')
const app = require('./package.json')
const fs = require('fs')
const { commit, config, slugify, login, logout, upload } = require('./lib/helpers')
const tar = require('tar')
const pack = readPackage(cwd)
const semver = require('semver')
const extractor = require('./lib/extract');

// Check on current directory first - if its a git and mpack directory
let isGit = false
git.silent(true).status(async (err, data) => {
    isGit = !err
    if (data) {
        const lastTag = await new Promise((done, fail) => git.raw(['describe', '--tags', '--abbrev=0' ], (err, data) => err ? done('') : done(data)))
        pack.last = lastTag.trim() || '4b825dc642cb6eb9a060e54bf8d69288fbee4904' // initial hash for 1st commit
        pack.gitPath = await new Promise((done, fail) =>
            git.revparse('--show-toplevel'
            , (err, data) => err ? done(err) : done(data)))
        if (pack.last) {
            const files = await new Promise((done, fail) => git.raw(['diff', '--name-only', pack.last ], (err, data) => err ? fail(err) : done(data))) || ''
            const pack_path = app.name ? `${app.name}/` : ''

            pack.files = files.split('\n').filter(f => f && (!pack_path || f.startsWith(pack_path)))
        }
        pack.ok = typeof pack.deps === 'object' && typeof pack.path === 'string'
        // const tags = await new Promise((done, fail) => git.raw(['log', '--tags', '--simplify-by-decoration', '--pretty=%h__%cI__%D', '-10' ], (err, data) => err ? fail(err) : done(data)))
        // Object.assign(gitData, data, { tags: tags.split('\n').map(l => {
        //     const [h, d, t] = l.split('__')
        //     , [ branch, tag ] = (t || '').split('tag: ')
        //     return { hash: h, date: new Date(d), branch: branch || '', tag: tag || '' }
        // }) })
    }
    start()
})


function start() {
    figlet('MPack', function(err, data) {
        if (err) return console.dir(err)
        console.log(green(data), '   ver.', app.version)
        console.log('')
        if (!pack.ok && isGit) console.log('  Tip: Go to the directory of pack for different menu')

        if (args._.length < 1 || !isGit) {
            return welcome()
        }
        run(args._[0])
    })
}

async function welcome() {
    // check if the current directory is a git
    // also if there is package.json
    // process.cwd()
    const choices = [ 'exit', 'extract' ]
    if (pack.ok) choices.push('check', 'make', 'release', 'update')
    else choices.push('new')
    choices.push('login', 'logout')

    const answer = await prompt([
        { name: 'task', message: `What do you want to do?`, type: 'list', choices }
    ])
    if (answer.task === 'exit') return console.log('👋 Bye!\n')
    // console.log('answer', answers)
    if (answer.task === 'extract') answer.task = 'extractMenu'
    run(answer.task)
}

async function run(task) {
    switch (task) {
        case 'login': doLogin(); break // test login to url and save config
        case 'logout': logout(); break // clears the config
        case 'check': console.log(pack); break
        case 'config': {
            console.log('Config:\n', config().all); break
        }
        case 'new': doNew(); break
        case 'make': doMake(); break
        case 'release': doMake(true); break
        case 'deps': doDeps(); break
        case 'extract': extractor.extract(); break // extract to yaml with crawl - Heng's functions
        case 'update':
        case 'upload': doUpload(); break
        case 'extractMenu': doExtract(); break
            // if tgz file exists
            // await doUpload(file, release)
    }
    // update and release will auto login first
}

async function doExtract() {
    const defaultConfig = require('./lib/extract-config.json');
    // development URL
    let {
        url = defaultConfig.url,
        id,
        username = defaultConfig.username,
        password = defaultConfig.password,
        folder,
        withJson
    } = require('minimist')(process.argv.slice(2));

    const ques = [
        { name: 'url', message: `Target URL`, default: url, type: 'input', validate: (d) => d.length > 0},
        { name: 'username', message: `Username`, default: username, type: 'input', validate: (d) => d.length > 0 },
        { name: 'password', message: `Password`, mask: '*', default: password, type: 'password', validate: (d) => d.length > 0 },
        { name: 'id', message: `IDs (* or comma-seperated)`, default: id, type: 'input', validate: (d) => d.length > 0 }
    ]
    let ans = await prompt(ques)

    let breakdownAns;
    if (ans.id === '*') {
        breakdownAns = await prompt([
            {name: 'type', message: `Choose type of items for extraction`, type: 'checkbox', choices: ['lookups', 'items'], validate: (d) => d.length > 0}
        ])

        if (breakdownAns.type.length && breakdownAns.type.indexOf('lookups') > -1) {
            lTypeAns = await prompt([
                {name: 'ltype', message: `Lookup Types: (comma-seperated)`, type: 'input', validate: (d) => d.length > 0}
            ])

            if (lTypeAns.ltype && lTypeAns.ltype.length) {
                const ltypes = lTypeAns.ltype.split(',');
                breakdownAns.type.splice(breakdownAns.type.indexOf('lookups'), 1);

                ltypes.forEach(ltype => {
                    breakdownAns.type.push(`lookups:${ltype}`)
                })
            }
        }
        
        ans = {
            ...ans,
            ...breakdownAns
        }
    }

    const remainingAns = await prompt([
        { name: 'folder', message: `Folder ID (${breakdownAns && breakdownAns.type.indexOf('items') > -1 ? 'Required' : 'Optional'})`
            , default: folder, type: 'input', validate: breakdownAns && breakdownAns.type.indexOf('items') > -1 ? (d) => d.length > 0 : undefined},
        { name: 'withJson', message: `With JSON File?`, default: withJson, type: 'confirm' }
    ])

    ans = {
        ...ans,
        ...remainingAns
    }

    extractor.extract(ans);
}

async function doNew() {
    // TODO: check if current or parent dir pack's package.json exist
    if (pack.ok) return console.log(`📦  Pack found here. Try change to parent directory\n`)
    const ques = [
        { name: 'name', message: `📦 Name your pack (min 5 chars)`, type: 'input', validate: (d) => d.length > 4 },
    ]
    if (!isGit) ques.push({ name: 'init', message: `${cwd} is not git repo, initialize?`, type: 'confirm' })
    const answer = await prompt(ques)
    //
    if (answer.name) {
        const name = slugify(answer.name)

        if (!isGit && answer.init) {
            await new Promise((done, fail) => git.init(false, (err, d) => { err ? fail(err) : done(d) }))
            console.log(`Intialized git in ${cwd}`)
            pack.gitPath = cwd
        }
        else if (!isGit) return

        const path = `${cwd}/${name}`
        if (fs.existsSync(path)) {
            // package.json found?
            if (fs.existsSync(`${path}/package.json`)) {
                console.log(redBright(`⚠️ Directory '${name}' already contains package.json`))
                return welcome()
            }
            // return console.log(redBright(`Directory '${name}' already exists`))
        }
        else fs.mkdirSync(path)

        setPackage({ name,
            version: "1.0.0",
            description: '',
            deps: {},
            path: `.${cwd.substr(pack.gitPath.length)}/${name}`,
            requires: {
                server: "",
                web: ""
            },
            perms: [],
            mpack: app.version
        }, path)
    console.log(green(`👍 Created directory for '${name}'\n`))
    process.chdir(name)
    }
}

async function doLogin() {
    // TODO: login to registry OR development
    const ques = [
        { name: 'url', message: `URL of Registry`,
            default: config().get('url') || 'https://hub.datumcorp.com',
            type: 'input', validate: (d) => d.startsWith('https://')
        },
        { name: 'key', message: `🔑 Login Key`, type: 'input' },
    ]
    const ans = await prompt(ques)
    if (ans.url && ans.key) {
        if (await login(ans.url, ans.key)) {
            console.log(green(`👍 Logged in to '${ans.url}'\n`))
            return true
        }
        else console.log(green(redBright`Login incorrect\n`))
    }
    else console.log('Aborted.')
    return false
}

async function doDeps() {
    // add / remove / update / check
    console.log(args)
}

async function doMake(release = false) {
    // list all files in the pack dir
    if (!pack.ok) return console.log(`⚠️  Pack not found. Are you in the correct directory?\n`)
    if (!pack.files) return console.log(`⚠️  Nothing to do! Go commit some files first.\n`)
    if (release) {
        // TODO: check if the git files committed
    }

    // update pack's package.json mpack
    const version = semver.inc(pack.version, release ? 'patch' : 'prerelease')
    setPackage({ mpack: app.version, version }, cwd)

    // needs to be in current dir
    pack.files.push('package.json')
    // console.log(pack.files)
    const file = `${pack.name}@${version}.tgz`
    await tar.c({ gzip: true, file }, pack.files)
    return file
}

async function doUpload(fn, release) {
    // if not logged in yet
    if (!config.get('token')) {
        if (!await doLogin()) return
    }

    upload(fn, { release })
    // upload pack
    // commit to git
}

// #region Helpers

function readPackage(path) {
    try {
        const p = require(`${path || cwd}/package.json`)
        return JSON.parse(JSON.stringify(p))
    }
    catch (e) {
        // console.error(e.message)
        return {}
    }
}

function setPackage(data, path) {
    const d = readPackage(path)
    Object.assign(d, data)
    fs.writeFileSync(`${path}/package.json`, JSON.stringify(d, null, 2))
}

// #endregion

