var request = require("request");
const sha256 = require('js-sha256');
const headers = {
    'Cache-Control': 'no-cache',
    'Content-Type': 'application/json'
};
let url = ''
function setUrl (targetUrl) {
    url = `${targetUrl}/api`;
    console.log('url is now:', url)
}

function login(username, password) {
    if (!username) { throw new Error('--username expected'); }
    if (!password) { throw new Error('--password expected'); }

    const hashedPw = sha256((username.toString() + password.toString()))
    const options = {
        method: 'POST',
        url: url + '/user/login',
        headers,
        body: {
            "username": username
            ,"password": hashedPw
        },
        json: true
    };

    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            headers.cookie = response.headers["set-cookie"]
            console.log('login')
            resolve(body);
        });
    });
}

function logout() {
    const options = {
        method: 'GET',
        url: url + '/user/logout',
        headers
    };

    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            console.log('logout')
            resolve(body);
        });
    });
}

const lookupCache = {};

function getLookup(id) {
    const options = {
        method: 'GET',
        url: url + '/lookup/' + id,
        headers
    };

    return new Promise((resolve, reject)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            
            let bodyBuffer = body;
            try {
                bodyBuffer = JSON.parse(bodyBuffer);
                if (!bodyBuffer.ok) { reject(bodyBuffer) }
            }
            catch (e) {
                console.error('e', e);
                reject(e)
            }
            finally {
                if (bodyBuffer.data) {
                    lookupCache[id] = bodyBuffer.data;
                    lookupCache[`${bodyBuffer.data.code}:${bodyBuffer.data.type}`] = bodyBuffer.data;
                resolve(bodyBuffer.data);
            }
            }
        });
    });
}

function getLookups(lookupType) {
    const options = {
        method: 'GET',
        url: url + '/lookups/' + lookupType + '?all=1',
        headers
    };

    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            
            let bodyBuffer = body;
            try {
                bodyBuffer = JSON.parse(bodyBuffer);
            }
            catch (e) {
                console.error('e', e);
            }
            finally {
                resolve(bodyBuffer.data);
            }
        });
    });
}

function getItem(id, folder) {
    const options = {
        method: 'GET',
        url: url + '/item/' + id + '?folderId=' + folder,
        headers
    };

    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            let bodyBuffer = body;
            try {
                bodyBuffer = JSON.parse(bodyBuffer);
            }
            catch (e) {
                console.error('e', e);
            }
            finally {
                resolve(bodyBuffer.data);
            }
        });
    });
}

function getItems(folder) {
    const options = {
        method: 'GET',
        url: url + '/items?limit=0&folderId=' + folder + '&find={"folder.id": "' + folder + '"}',
        headers
    };

    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            let bodyBuffer = body;
            try {
                bodyBuffer = JSON.parse(bodyBuffer);
            }
            catch (e) {
                console.error('e', e);
            }
            finally {
                resolve(bodyBuffer.data);
            }
        });
    });
}

function getFolderSettings(folderId, keys) {
    const specificHeaders = {...headers};
    specificHeaders.folderId = folderId;

    const options = {
        method: 'GET',
        url: url + `/folder/${folderId}/settings/${keys.toString()}`,
        headers: specificHeaders
    };

    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            
            let bodyBuffer = body;
            try {
                bodyBuffer = JSON.parse(bodyBuffer);
            }
            catch (e) {
                console.error('e', e);
            }
            finally {
                resolve(bodyBuffer.data);
            }
        });
    });
}

function getSchedPrefs(docid, folderId) {
    const find = JSON.stringify({"docid": docid});
    const specificHeaders = {...headers};
    specificHeaders.folderId = folderId;

    const options = {
        method: 'GET',
        url: url + '/schedules?find=' + encodeURIComponent(find),
        headers: specificHeaders
    };

    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            
            let bodyBuffer = body;
            try {
                bodyBuffer = JSON.parse(bodyBuffer);
            }
            catch (e) {
                console.error('e', e);
            }
            finally {
                resolve(bodyBuffer.data);
            }
        });
    });
}

function getLookupByCodeAndType(code, type) {
    if (lookupCache[`${code}:${type}`]) {
        throw new Error('SKIP');
        // console.log(`GET lookup ${code}:${type} FROM CACHE`);
        // return Promise.resolve(lookupCache[`${code}:${type}`])
    }
    console.log(`GET lookup ${code}:${type}`);

    const options = {
        method: 'GET',
        url: `${url}/lookups/${type}?find={"code":["${code}"]}`,
        headers
    };
    return new Promise((resolve)=>{
        request(options, function (error, response, body) {
            if (error) throw new Error(error);
            let bodyBuffer = body;
            try {
                bodyBuffer = JSON.parse(bodyBuffer);
            }
            catch (e) {
                console.error('e', e);
            }
            finally {
                lookupCache[bodyBuffer.data[0].id] = bodyBuffer.data[0];
                lookupCache[`${bodyBuffer.data[0].code}:${bodyBuffer.data[0].type}`] = bodyBuffer.data[0];
                resolve(bodyBuffer.data[0]);
            }
        });
    });
}

class Lookups {
    static async extract(lookupType, folder) {
        const lookups = await getLookups(lookupType);
        let finalArray = [];
        
        for (const lookup of lookups) {
            try {
                if (
                    !lookupCache[lookup.id]
                ) {
                    let lookupArray = await Lookup.getDependencies(lookup, folder);
                    
                    lookupCache[lookup.id] = lookup;
                    lookupCache[`${lookup.code}:${lookup.type}`] = lookup;
                    
                finalArray = finalArray.concat(lookupArray);
            }
            }
            catch (error) {
                console.log('error', error.message);
            }
        }
        console.log('Lookups - finalArray - length:', finalArray.length);
        return finalArray;
    }
}

class Lookup {
    constructor(id) {
        this.id = id;
    }

    extract() {
        if (lookupCache[this.id]) {
            throw new Error('SKIP');
            // console.log(`GET lookup ${this.id} FROM CACHE`);
            // return Promise.resolve(lookupCache[this.id])
        }
        else {
            console.log(`GET lookup ${this.id}`);
        return getLookup(this.id);
    }
    }

    static async getDependencies(lookup, folder) {
        console.log(`> L${lookup.type}:${lookup.id}`);
        switch (lookup.type) {
            case 8: return await (new L8(lookup.id)).handle(lookup);
            case 9: return await (new L9(lookup.id)).handle(lookup);
            case 13: return await (new L13(lookup.id, folder)).handle(lookup);
            case 90: return await (new L90(lookup.id, folder)).handle(lookup);
            case 40: return await (new L40(lookup.id)).handle(lookup);
            default: return [lookupArray];
        };
    }
}

class L13 extends Lookup {
    constructor(id, folder) {
        super(id);
        this.folder = folder;
    }

    async extract(showLength) {
        process.stdout.write('L13: ');
        let l13
        try {
            l13 = await super.extract()
        }
        catch (error) {
            if (error && error.message !== 'SKIP') {
                process.stdout.write('L13: FALLBACK: ');
                l13 = await getLookupByCodeAndType(this.id, 13)
            }
            else {
                console.log('SKIPPED:', this.id);
                throw new Error('SKIP');
            }
        }

        return await this.handle(l13, showLength);
    }

    async handle(l13, showLength) {
        const extractedSources = [];
        let { sources = [] } = l13.data;
        let toReturn = [l13];

        if (!sources || !Array.isArray(sources)) { sources = [] }

        for (const source of sources) {
            if (typeof source.ds === 'string' && source.ds.trim() && extractedSources.indexOf(source.ds.trim()) < 0) {
                try {
                    const dependencyLookup = new L9(source.ds);
                    const dependencyLookupArray = await dependencyLookup.extract(); 
                    toReturn = toReturn.concat(dependencyLookupArray || []);
                }
                catch (error) {
                    if (error && error.message !== 'SKIP') {
                        console.log('error:', error);
                }
                    else {
                        console.log('SKIPPED:', source.ds);
                    }
                }
                extractedSources.push(source.ds.trim())
            }
        }

        if (this.folder && l13.val === 3) {
            const fsetting = await getFolderSettings(this.folder, [`x-${l13.code}`]);
            fsetting._APP = 'settings';
            toReturn.push(fsetting);
        }

        if (showLength) {
            console.log(`L13:${this.id}->length of toReturn:`, toReturn.length)
        }
        
        return toReturn;
    }
}

class L9 extends Lookup {
    constructor(id) {
        super(id);
    }

    async extract(showLength) {
        process.stdout.write('L9: ');
        let l9
        try {
            l9 = await super.extract()
        }
        catch (error) {
            if (error && error.message !== 'SKIP') {
                process.stdout.write('L9: FALLBACK: ');
            l9 = await getLookupByCodeAndType(this.id, 9)
        }
            else {
                console.log('SKIPPED:', this.id);
                throw new Error('SKIP');
            }
        }
        
        return await this.handle(l9, showLength)
    }

    async handle(l9, showLength) {
        const {forms = '', findForm = '', settingKeys} = l9.data
            , settingDependency = settingKeys ? settingKeys.split('x-')[1] : ''
            , formDependency = [(forms || '').trim(), (findForm || '').trim()].filter(lkId => lkId)
            , extractedFormDependencies = [];
        let toReturn = [l9];

        if (settingDependency) {
            try {
                const dependencyLookup = new L13(settingDependency)
            const dependencyLookupArray = await dependencyLookup.extract()
            toReturn = toReturn.concat(dependencyLookupArray || []);
        }
            catch (error) {
                if (error && error.message !== 'SKIP') {
                    console.log('error:', error);
        }
                else {
                    console.log('SKIPPED:', settingDependency);
                }
            }
        }

        for (const lkId of formDependency) {
            if (extractedFormDependencies.indexOf(lkId) < 0) {
                try {
                    const dependencyLookup = new L13(lkId)
                const dependencyLookupArray = await dependencyLookup.extract()
                toReturn = toReturn.concat(dependencyLookupArray || []);
                }
                catch (error) {
                    if (error && error.message !== 'SKIP') {
                        console.log('error:', error);
                    }    
                    else {
                        console.log('SKIPPED:', lkId);
                    }
                }
                extractedFormDependencies.push(lkId)
            }
        }

        if (showLength) {
            console.log(`L9:${this.id}->length of toReturn:`, toReturn.length)
        }

        return toReturn;
    }
}

class L40 extends Lookup {
    constructor(id) {
        super(id);
    }

    async extract(showLength) {
        process.stdout.write('L40: ');
        const l40 = await super.extract();
        return await this.handle(l40, showLength)
    }

    async handle(l40, showLength) {
        const { layout = [] } = l40.data;
        let toReturn = [l40];

        if (layout[0] && layout[0][2] && typeof layout[0][2] === 'string' && layout[0][2].trim()) {
            try {
            const dependencyLookup = new L9(layout[0][2]);
            const dependencyLookupArray = await dependencyLookup.extract();
            toReturn = toReturn.concat(dependencyLookupArray || []);
        }
            catch (error) {
                if (error && error.message !== 'SKIP') {
                    console.log('error:', error);
                }
                else {
                    console.log('SKIPPED:', layout[0][2]);
        }
            }
        }
        
        if (showLength) {
            console.log(`L40:${this.id}->length of toReturn:`, toReturn.length);
        }

        return toReturn;
    }
}

class L90 extends Lookup {
    constructor(id, folder) {
        super(id);
        this.folder = folder
    }

    async extract(showLength) {
        process.stdout.write('L90: ');
        const l90 = await super.extract();
        return await handle(l90, showLength);
    }

    async handle(l90, showLength) {
        const { dsid, reports } = l90.data;
        let toReturn = [l90];

        if (dsid && dsid.trim()) {
            const dependencyLookup = new L9(dsid);
            try {
            const dependencyLookupArray = await dependencyLookup.extract();
            toReturn = toReturn.concat(dependencyLookupArray || []);
        }
            catch (error) {
                if (error && error.message !== 'SKIP') {
                    console.log('error:', error);
                }
                else {
                    console.log('SKIPPED:', dsid);
                }
            }
        }

        if (reports && reports.length) {
            const reportIds = reports.filter(report => report && report.id && report.id.trim()).map(report => report.id)

            for (const l40Id of reportIds) {
                const dependencyLookup = new L40(l40Id);
                try {
                const dependencyLookupArray = await dependencyLookup.extract();
                toReturn = toReturn.concat(dependencyLookupArray || []);
            }
                catch (error) {
                    if (error && error.message !== 'SKIP') {
                        console.log('error:', error);
                    }
                    else {
                        console.log('SKIPPED:', l40Id);
        }
        }
            }
        }

        if (this.folder) {
            try {
            const schedPrefsAndFolderSettings = await (new SchedPrefs(this.id, this.folder)).extract()
            const {schedPrefs, folderSettings} = schedPrefsAndFolderSettings;
                console.log(`L90:${this.id}->length of schedPrefs:`, schedPrefs.length);

            toReturn = toReturn.concat(
                (schedPrefs || []).map(schedPref => {
                    schedPref.APP = 'SCH';
                    return schedPref;
                })
            ).concat(folderSettings);
        }
            catch (e) { console.error(e); }
        }

        if (showLength) {
            console.log(`L90:${this.id}->length of toReturn:`, toReturn.length);
        }

        return toReturn;
    }
}

class L8 extends Lookup {
    constructor(id, folder) {
        super(id);
        this.folder = folder
    }

    async extract(showLength) {
        process.stdout.write('L8: ');
        const l8 = await super.extract();
        return await this.handle(l8, showLength)
    }

    async handle(l8, showLength) {
        let toReturn = [l8];

        if (this.folder) {
            try {
            const schedPrefsAndFolderSettings = (new SchedPrefs(this.id, this.folder)).extract()
            const {schedPrefs, folderSettings} = schedPrefsAndFolderSettings;
                console.log(`L8:${this.id}->length of schedPrefs:`, schedPrefs.length);
            toReturn = toReturn.concat(
                (schedPrefs || []).map(schedPref => {
                    schedPref.APP = 'SCH';
                    return schedPref;
                })
            ).concat(folderSettings);
        }
            catch (e) { console.error(e); }
        }

        if (showLength) {
            console.log(`L8:${this.id}->length of toReturn:`, toReturn.length);
        }

        return toReturn;
    }
}

class SchedPrefs {
    constructor(docId, folderId) {
        this.docId = docId;
        this.folderId = folderId;
    }

    async extract() {
        const schedPrefs = (await getSchedPrefs(this.docId, this.folderId)) || [];

        const savedFSettings = []
        for(const schedPref of schedPrefs) {
            const folders = ((schedPref.data || {}).folders || []);
            
            for (const folderId of folders) {
                const fsetting = await getFolderSettings(folderId, ['tz']);
                fsetting._APP = 'settings';
                savedFSettings.push(fsetting);
            }
        }

        return {schedPrefs, folderSettings: savedFSettings}
    }
}

class Item {
    constructor(id, folder) {
        this.id = id;
        this.folder = folder;
    }

    async extract() {
        console.log('extracting Item:', this.id);
        const item = await getItem(this.id, this.folder);
        item.APP = 'ITEM'

        return item;
    }
}

class Items {
    static async extract(folder) {
        console.log('extracting Items');
        const items = await getItems(folder);
        
        return items.map(item => {
            item.APP = 'ITEM'
            return item
        });
    }
}

module.exports = {
    setUrl,
    getLookup,
    getLookupByCodeAndType,
    Lookup,
    Lookups,
    L13,
    L9,
    L8,
    L90,
    L40,
    Item,
    Items,
    login,
    logout
};
